# This is the file for backend-independent (only matplotlib) part of the GUI
# Isn't working yet
#
# I think I should redesign it. It should work from normal pylab
# session. So, it's just a stub for future work.

from pylab import * # TODO: import only needed stuff
from swan.utils import *

from matplotlib.figure import Figure

from swan import pycwt

class SwanGUI:
    def __init__(self,fname=None,
                 Fs = 2.0,
                 f0 = 1.5,
                 labelsize = 10.0,
                 verbose = 1):
        
        self.verbose = verbose
        self.deoutbf = True # wether to remove outbursts

        self.data = None
        self.Fs = Fs # sampling frequency

        self.rc = {'label-size': labelsize,
                   'wavelet' : 'Morlet',
                   }

        self.fig = Figure(figsize=(8,6), dpi=100)

        self.f_pars = {'start': 0.1, 'step': 0.025, 'stop': 10.,
                       'log?': False} #frequency parameters #self.set_freqs =

        self.skeletons = []
        self.skel_plots = []
        
        self.rind = -1

        self.cwt_freshly_runf = False
        self.ehndl = None # mouse click event handler
        self.cwt = None
        self.show_how = self.show_surface
        self.wavelet = pycwt.Morlet(f0)
        self.cwt_image = None

        # -- Prepare different parts --
        set_rgbswan_cm()
        self.setup_show_funcs()
        self.setup_axes()

        # -- Try to load data if the file name is provided --
        if defined(fname):
            self.load_data(fname)
            self.plot_data()

        self.window.show_all()

    def setup_show_funcs(self):
        """ Prepare functions for different show options """
        self.show_funcs = {
            'Modulus': lambda x: abs(x),
            'Phase': lambda x: arctan2(x.imag, x.real),
            'Real' : lambda x: x.real,
            'Imaginary': lambda x: x.imag,
            'EDS': self.eds,
            #'EDS': lambda x: (x.real**2. + x.imag**2.)/self.wavelet.fc,
            'EDS-A-1': lambda x: self.eds_a_helper(x,1),
            'EDS-A-2': lambda x: self.eds_a_helper(x,2)
                }
        self.image_logscale = False
        self.show_func = self.show_funcs['EDS'] # default view

    def eds(self, x):
        return pycwt.eds(x)/std(self.data)**2

    def eds_a_helper(self, x, alpha):
        """ Show EDS with ridge amplitude correction """
        #alpha = 1.9798196483 # magic alpha
        coefs = (self.freqs**alpha) * 2.0/ pi
        eds = self.eds(x)
        if len(eds.shape) >1: 
            for i in xrange(eds.shape[1]):
                eds[:,i] *= coefs
        else:
            eds *= coefs
        #return eds/self.wavelet.fc
        return eds
    
    def load_data(self,fname):
        """ Loads the data """
        try:
            data = numpy.loadtxt(fname)
            s = data.shape
            self.fname = fname
            
            if len(s) < 2:
                if len(data)%2 :
                    data = data[:-1]
                    #self.data = data - mean(data)
                if self.deoutbf:
                    data = deoutburst(data)
                self.data = detrend1(data)

                self.time_from_Fs(self.Fs)
            else:
                if s[0] < s[1]:
                    data = transpose(data)
                    s = data.shape

                time = data[:,0] / 1000.0 # Here I assume that time is in ms
                data = data[:,1]
                
                if len(data)%2 :
                    data = data[:-1]
                    time = time[:-1]
                
                self.data = data
                self.time = time
                self.Fs = 1./mean(time[1:]- time[:-1])
        except:
            print "Error in data loading:"
            print sys.exc_type, sys.exc_value
            self.fname = None
            self.data = None

    def setup_axes(self):
        """Prepares axes for further work"""
        rc('xtick', labelsize = self.rc['label-size'])
        rc('ytick', labelsize = self.rc['label-size'])
        self.fig.clear()
        self.ax = [self.fig.add_axes((0.04, 0.2, 0.8, 0.78))]
        self.ax.append(self.fig.add_axes((0.04, 0.04, 0.8, 0.12),
                                         sharex=self.ax[0]))
        self.ax.append(self.fig.add_axes((0.85, 0.2, 0.04, 0.78), 
                                         xticklabels=[], 
                                         yticklabels=[],
                                         visible = False))
        self.ax[0].hold(True)
        self.ax[1].hold(False)


    def image_logscale_toggle(self,w):
        self.image_logscale = w.get_active();
        self.change_display(self['display_combo'])

    @with_data
    def on_apply_cwt_clicked(self, w):
        self.freqs = self.set_freqs()
        self.cwt_calc()
        self.cwt_display()


    def set_freqs(self):
        if self.f_pars['log?']:
            a = log2(self.f_pars['start'])
            b = log2(self.f_pars['stop'])
            if self.verbose:
                print a,b, 1/self.f_pars['step']
            return 2.**arange(a, b, self.f_pars['step'])
        else:
            return arange(self.f_pars['start'],
                          self.f_pars['stop'],
                          self.f_pars['step'])

        return 2.**arange(a, b, self.f_pars['step'])

    @with_data
    def cwt_calc(self):
        self.cwt = pycwt.cwt_f(self.data, self.freqs, Fs=self.Fs, \
                wavelet = self.wavelet)
        self.cwt_freshly_runf = True
        pass

    @with_cwt
    def cwt_display(self):
        """Show the results of the wavelet transform"""
        
        self.clear_all_rhythms()
        self.ax[0].clear()
        self.rm_skeletons()
        self['local_maxima_chk'].set_active(False)

        self.show_how()
        self.show_cone()
        setp(self.ax[0],'yscale',('linear','log')[self.f_pars['log?']])
        self.show_colorbar()

    @with_cwt
    def show_cone(self):
        w = self.wavelet
        f = self.freqs
        ax = self.ax[0]
        t = self.time[-1]
        try:
            ax.fill_betweenx(f, 0.0,
                             w.cone(f),
                             alpha=0.5, color='black')
            ax.fill_betweenx(f,
                             t+w.cone(-f),
                             t,
                             alpha=0.5, color='black')
        except:
            print("Can't use fill_betweenx function: update\
            maptlotlib?")


    
    
    @with_cwt
    def show_surface(self):
        ax = self.ax[0]
        self.cwt_image = ax.imshow(self.show_func(self.cwt),
                                   origin='lower',
                                   extent=(self.time[0], self.time[-1],
                                           self.freqs[0], self.freqs[-1]),
                                   interpolation='bilinear',
                                   cmap=self.curr_cmap)



    @with_cwt
    def show_contour(self):
        ax = self.ax[0]
        self.cwt_image = ax.contour(self.show_func(self.cwt),
                                    origin='lower',
                                    extent=(self.time[0], self.time[-1],
                                            self.freqs[0], self.freqs[-1]),
                                    cmap=self.curr_cmap)
    @with_cwt
    def show_colorbar(self):
        self.fig.colorbar(self.cwt_image, cax=self.ax[2])
        setp(self.ax[2], visible=True)

    @with_data
    def time_from_Fs(self,Fs=2.0):
        tfinish = len(self.data)/Fs
        self.time = arange(0,1.2*tfinish,1./Fs)[:len(self.data)]


    @with_data
    def plot_data(self):
        """Plot current data"""
        if self.time is None:
            self.ax[1].plot(self.data,'k-')
        else:
            self.ax[1].plot(self.time,self.data,'k-')
            setp(self.ax[1],'xlim',(self.time[0], self.time[-1]))
        self.canvas.draw()
        return 0

    def renew_skelsp(self):
        return defined(self.data) and \
                (len(self.skeletons) is 0) and\
                self.cwt_freshly_runf

    
    def plot_a_skeleton(self,skel):
        return self.ax[0].plot(skel[:,0], skel[:,1], 'y.', markersize = 0.1)

    def set_ax0_tight(self):
        setp(self.ax[0], 'xlim', (0, self.time[-1]),\
                'ylim', (self.freqs.min(), self.freqs.max()))
    
    def skeletons_shade(self, flag=True):
        for plh in self.skel_plots:
            setp(plh,'visible', not flag)
        self.canvas.draw()

    def calc_skeletons(self):
        """calculate local modulus maxima (skeleton lines)"""
        L =  len(self.data)
        del(self.skeletons[:])
        for i in xrange(L):
            xe,ye = get_extr_deriv(self.freqs, self.show_func(self.cwt[:,i]))
            ye[:] = self.time[i]
            self.skeletons.append(transpose(array([ye,xe])))
            if not i%10:
                sys.stderr.write (make_progress_str(i,L))
    
#------------------- End of the SwanGUI class ------------




# Colormaps for Swan

#from numpy import *
from pylab import *
import matplotlib


def hsv2rgb(ctuple):
    """Convert hsv tuple to rgb tuple
    based on wikipedia page"""
    h,s,v = ctuple
    Hi = mod(floor(h/60.), 6)
    f = h/60. - floor(h/60.)
    p = v*(1 - s)
    q = v*(1 - s*f)
    t = v*(1 - (1-f)*s)
    
    if (Hi == 0): r,g,b = v,t,p
    elif (Hi == 1): r,g,b = q,v,p
    elif (Hi == 2): r,g,b = p,v,t
    elif (Hi == 3): r,g,b = p,q,v
    elif (Hi == 4): r,g,b = t,p,v
    elif (Hi == 5): r,g,b = v,p,q
    
    return r,g,b

def get_rgbswan_data1():
    return {'red':   ((0, 0, 0), 
                      (0.2, .1, .1),
                      (0.4, .4, .4),
                      (0.6, .6, .6),
                      (0.8, 1, 1),
                      (1, 1, 1)),
            'green': ((0., 0, 0), 
                      (0.2, 0.5, 0.5), 
                      (0.4, 1, 1),
                      (0.6, .8, .8),
                      (0.8, 0.9, 0.9),
                      (1, 0.2, 0.2)),
            'blue':  ((0, 1, 1), 
                      (0.2, 0.6, 0.6), 
                      (0.4, 0.3, 0.3),
                      (0.6, 0, 0),
                      (0.8, 0, 0),
                      (1.0, 0, 0))}

def get_rgbswan_data2():
    sat,val = 0.9,0.95
    colors = {'red':0,'green':1,'blue':2}
    step = 0.05

    rgb_vect = [ (x, hsv2rgb(((1-x)*220, sat, val))) 
                 for x in arange(0,1 + step,step)]
    res = {}

    for color,ind in colors.items():
        res[color] = [(x[0], x[1][ind], x[1][ind]) for x in rgb_vect]    
    return res

def set_rgbswan_cm():
    LUTSIZE = mpl.rcParams['image.lut']
    
    _rgbswan_data =  get_rgbswan_data2() 

    cm.datad['rgbswan'] = _rgbswan_data
    cm.rgbswan  = matplotlib.colors.LinearSegmentedColormap('rgbswan',
                                                            _rgbswan_data, LUTSIZE)
    return cm.rgbswan



def rgbswan():
    '''
    set the default colormap to jet and apply to current image if any.
    See help(colormaps) for more information
    '''
    rc('image', cmap='rgbswan')
    im = gci()

    if im is not None:
        im.set_cmap(cm.rgbswan)
    draw_if_interactive()



import gtk
from pylab import setp, arange, array

def best (scoref, lst):
    n,winner = 0, lst[0]
    for i, item in enumerate(lst):
        if  scoref(item, winner): n, winner = i, item
    return n,winner

def min1(scoref, lst):
    return best(lambda x,y: x < y, map(scoref, lst))

def in_range(n, region):
    return (n > region[0]) and (n < region[1])

def nearest_item_ind(items, x, fn = lambda a: a):
    return min1(lambda p: abs(fn(p) - x), items)[0]

class banded_point:
    """2D point with a +- band attached to it"""
    def __init__(self, x, y, band):
        self.x = x
        self.y = y
        self.b = band
        

class Rhythm_path:
    """ Add rhythm path to the canvas """
    def __init__(self, canvas, i=0, band=0.5):
        self.x = []   # x of the skeleton line
        self.y = []   # y of the skeleton line
        self.v = []   # indices of the eds
        self.px = []  # xpoints in path
        self.py = []  # ypoints in path
        self.bpoints  = [] # banded points in path
        #---------------------------------
        self.band = band      # Frequency band (herz)
        self.plhndl = []     # Plot handle
        #---------------------------------
        self.canvas = canvas
        self.id = i
        self.butcb = [None,None,None] # Buttons callbacks
        self.box = gtk.HBox(False,0)
        self.b_adj = gtk.Adjustment(value=band,
                                    lower=0.0,
                                    upper=1000,
                                    step_incr=0.01,
                                    page_incr=0.04)
        band_spin = gtk.SpinButton(self.b_adj,climb_rate=0.2,digits=3)
        band_spin.set_numeric(True)
        band_spin.connect('value-changed',self.cb_band)
        self.buts = [gtk.ToggleButton('%d' %int(i)),
                     gtk.CheckButton('P'), gtk.CheckButton('R'),
                     band_spin]
        for b in self.buts:
            self.box.pack_start(b,False,False,0)

    def __del__(self):
        print 'Destructor of RhythmPath'
        for b in self.buts:
            self.box.remove(b)
        self.clear()
        self.rclear()
        for i in xrange(4):
            setp(self.plhndl[i], 'data', (self.px, self.py), 'visible', False)
        self.canvas.draw()
        pass
    
    def push_point(self, x, y):
        self.bpoints.append(banded_point(x, y, self.band))
        if len(self.bpoints) > 1:
            self.bpoints.sort(lambda p1,p2: cmp(p1.x, p2.x))

    def pop_point(self, x, y):
        n = nearest_item_ind(self.bpoints, x, lambda p: p.x)
        self.bpoints.pop(n)

    def clear(self):
        self.bpoints = []
    
    def rclear(self):
        """ Rhythm clear """
        while len(self.x) > 0:
            self.x.pop()
            self.y.pop()
            self.v.pop()

    def plot_setup0(self,ax,L):
        self.plhndl.append(ax.plot(self.px,self.py,'or-'))
        self.plhndl.append(ax.plot(self.px,array(self.py)+self.band,'y-'))
        self.plhndl.append(ax.plot(self.px,array(self.py)-self.band,'y-'))
        self.plhndl.append(ax.plot(self.x, self.y,'-',color='w',linewidth=5.0, alpha=0.8))
        setp(ax,'xlim',L[0], 'ylim',L[1])

    def vectors_from_bpoints(self):
        px = array(map(lambda p: p.x, self.bpoints))
        py = array(map(lambda p: p.y, self.bpoints))
        pl = py - self.band
        pu = py + self.band
        return px,py,pu,pl

    def plot_setup(self,ax,L):
        
        px,py,pu,pl = self.vectors_from_bpoints()
        
        self.plhndl.append(ax.plot(px, py, 'or-'))
        self.plhndl.append(ax.plot(px, pu, 'y-'))
        self.plhndl.append(ax.plot(px, pl, 'y-'))
        self.plhndl.append(ax.plot(self.x, self.y,'-',color='w',linewidth=5.0, alpha=0.5))
        setp(ax,'xlim',L[0], 'ylim',L[1])

    def preplot(self,canvas,vis=True):
        
        px,py,pu,pl = self.vectors_from_bpoints()

        setp(self.plhndl[0],'data', (px, py),'visible',vis)
        setp(self.plhndl[1],'data', (px, pu),'visible',vis)
        setp(self.plhndl[2],'data', (px, pl),'visible',vis)
        canvas.draw()

    def cb_band(self,w,data=0):
        """Callback for changing the frequency band for rhythm selection"""
        self.band = w.get_value()
        self.preplot(self.canvas)

    def rreplot(self,canvas,vis=True):
        setp(self.plhndl[3],'data', (self.x,self.y),'visible',vis)
        canvas.draw()





# General purpose utilities

from pylab import *

def log2(x):
    """Base 2 logarithm"""
    return log(x)/log(2.)

def make_progress_str(i,L=100):
    """Make a progress string with a rotating stick"""
    return '\r%3d%s %s' % (100*i/L, '%', ('|','/','-','\\')[i%4])


def deriv(x,y):
    """Calculate derivative of a signal"""
    return  .5*((y[1:-1] - y[0:-2]) / (x[1:-1] - x[0:-2]) + \
            (y[2:]   - y[1:-1]) / (x[2:]   - x[1:-1]))

def get_extr_deriv(x,y,maxf=True,wth=4):
    """Derivative-based extrema detection"""
    der = deriv(x,y)
    a = der > 0
    if maxf:
        a = a[1:] < a[:-1]
    else:
        a = a[1:] > a[:-1]

    k = arange(len(a))[a]
    xe = ones(len(k),dtype='d')
    ye = ones(len(k),dtype='d')

    for j,i in enumerate(k):
        p = polyfit(x[i+1:i+3], der[i:i+2], 1)
        xe[j] = -p[1] / p[0] #???
        p = polyfit(x[i:i+3], y[i:i+3], 2)
        #ye[j] = polyval(p, x[i+1:i+2])[0]
        ye[j] = polyval(p, (xe[j],) )[0]
    return xe,ye

def detrend1(vect, n=1):
    j = arange(len(vect))
    return vect - polyval(polyfit(j, vect, n), j)

def deoutburst(vect, n = 4):
    d = std(vect)
    m = mean(vect)

    low = m - n*d
    high = m+ n*d

    res = vect.copy()
    for i in xrange(len(vect)):
        if vect[i] < low:
            res[i] = low
        elif vect[i] > high:
            res[i] = high
        else:
            res[i] = vect[i]
    return res




# Continuous wavelet transfrom via Fourier transform
# Collection of routines for wavelet transform via FFT algorithm


#-- Some naming and other conventions --
# use f instead of omega wherever rational/possible
# *_ft means Fourier transform

#-- Some references --
# [1] Mallat, S.  A wavelet tour of signal processing
# [2] Addison, Paul S. The illustrated wavelet transform handbook
# [3] Torrence and Compo. A practical guide to wavelet
#     analysis. Bulletin of American Meteorological Society. 1998, 79(1):61-78

import numpy
from numpy.fft import fft, ifft, fftfreq

#try:
#    from scipy.special import gamma
#except:



pi = numpy.pi

class DOG:
    """Derivative of Gaussian, general form"""
    # Incomplete, as the general form of the mother wavelet
    # would require symbolic differentiation.
    # Should be enough for the CWT computation, though

    def __init__(self, m = 1.):
        self.order = m
        self.fc = (m+.5)**.5 / (2*pi)
    
    def psi_ft(self, f):
        c = 1j**self.order / numpy.sqrt(gamma(self.order + .5)) #normalization
        w = 2*pi*f
        return c * w**self.order * numpy.exp(-.5*w**2)

class Mexican_hat:
    def __init__(self, sigma = 1.0):
        self.sigma = sigma
        self.fc = .5 * 2.5**.5 / pi
    def psi_ft(self, f):
        """Fourier transform of the Mexican hat wavelet"""
        c = numpy.sqrt(8./3.) * pi**.25 * self.sigma**2.5 
        wsq = (2. * pi * f)**2.
        return -c * wsq * numpy.exp(-.5 * wsq * self.sigma**2.)
    def psi(self, tau):
        """Mexian hat wavelet as described in [1]"""
        xsq = (tau / self.sigma)**2.
        c = 2 * pi**-.25 / numpy.sqrt(3 * self.sigma) # normalization constant from [1]
        return c * (1 - xsq) * numpy.exp(-.5 * xsq)
    def cone(self, f):
        "e-folding time [Torrence&Compo]. For frequencies"
        return self.f0*2.0**0.5 / f
    def cone_s(self, s):
        "e-folding time [Torrence&Compo]. For scales"
        return 2.**0.5*s
    def set_f0(self, f0):
        pass

def heavyside(x):
    return 1.0*(x>0.0)

class Morlet:
    def __init__(self, f0 = 1.5):
        self.set_f0(f0)
    def psi(self, t):
        pi**0.25 * exp(-t**2 / 2.) * exp(2j*pi*self.fc*t) #[3]
    def psi_ft(self, f):
        """Fourier transform of the approximate Morlet wavelet
            f0 should be more than 0.8 for this function to be
            correct."""
        # [3]
        coef = (pi**-.25) * heavyside(f)
        return  coef * numpy.exp(-.5 * (2. * pi * (f - self.fc))**2.)
    def cone(self, f):
        "e-folding time [Torrence&Compo]. For frequencies"
        return self.f0*2.0**0.5 / f
    def cone_s(self, s):
        "e-folding time [Torrence&Compo]. For scales"
        return 2.**0.5*s
    def set_f0(self, f0):
        self.f0 = f0
        self.fc = f0


def next_pow2(x):
    return 2.**numpy.ceil(numpy.log2(x))

def pad_func(ppd):
    func = None
    if ppd == 'zpd':
        func = lambda x,a:  x*0.0
    elif ppd == 'cpd':
        func = lambda x,a:  x*a
    return func
    
def evenp(n):
    return not n%2

def cwt_a(signal, scales, sampling_scale = 1.0,
          wavelet=Mexican_hat(),
          ppd = 'zpd'):
    """ Continuous wavelet transform via fft. Scales version."""

    siglen = len(signal)

    if not evenp(siglen) :
        siglen -= 1
        signal = signal[:-1]

    needpad = wavelet.cone_s(numpy.max(scales))
    ftlen = int(next_pow2(siglen + 2*needpad))
    padlen = int((ftlen - siglen)/2)



    padseq  = numpy.ones((padlen,))
    padfunc = pad_func(ppd)


    padded_signal = numpy.concatenate((padfunc(padseq, signal[0]),
                                       signal,
                                       padfunc(padseq, signal[-1])))

    signal_ft = fft(padded_signal)     # FFT of the signal

    ## create the result matrix beforehand
    W = numpy.zeros((len(scales), siglen), 'complex')

    ftfreqs = fftfreq(ftlen, sampling_scale)  # FFT frequencies

    # Now fill in the matrix
    for n,s in enumerate(scales):
        psi_ft_bar = numpy.conjugate(wavelet.psi_ft(s * ftfreqs))
        psi_ft_bar *= (2*pi*s/sampling_scale)**0.5 # Normalization from [3]
        #W[n,:] = (s**.5) * ifft(signal_ft * psi_ft_bar)
        x = ifft(signal_ft * psi_ft_bar)
        
        W[n,:] = x[padlen:-padlen]
    return W


def cwt_f(signal, freqs, Fs=1.0, wavelet = Morlet(), ppd = 'zpd'):
    """Continuous wavelet transform -- frequencies version"""
    scales = wavelet.fc/freqs
    dt = 1./Fs
    return cwt_a(signal, scales, dt, wavelet, ppd)


def eds(x, f0=1.5):
    "Energy density surface [2,3]"
    ## Update, simulations with MK (as in [3]) suggest that I
    ## shouldn't divide by f0 to obtain correct normalisation,
    ## e.g. 1 s.d. in mean for white noise signal
    #return abs(x)**2/f0
    return abs(x)**2


def xwt_f(sig1, sig2, freqs, Fs=1.0, wavelet=Morlet()):
    "Cross-wavelet coeficients for 2 signals"
    cwtf = lambda x: cwt_f(x, freqs, Fs, wavelet)
    return xwt(cwtf(sig1),cwtf(sig2))

def absxwt_f(sig1, sig2, freqs, Fs=1.0, wavelet=Morlet()):
    "Cross-wavelet power for 2 signals"
    return abs(xwt_f(sig1,sig2, freqs, Fs, wavelet))/wavelet.f0**0.5

def absxwt_a(sig1, sig2, scales, dt=1.0, wavelet=Morlet()):
    "Cross-wavelet power for 2 signals"
    return abs(xwt_f(sig1,sig2, freqs, Fs, wavelet))/wavelet.f0**0.5


def xwt(wcoefs1,wcoefs2):
    "Cross wavelet transform for 2 sets of coefficients"
    return wcoefs1*wcoefs2.conjugate()

def absxwt(wcoefs1,wcoefs2, f0=1.5):
    "Cross-wavelet power for 2 sets of coefficients"
    ## Why do I divide by f_0^2 here?
    return  abs(xwt(wcoefs1,wcoefs2))/f0**0.5

def wtc_a(sig1, sig2, scales, dt=1.0, wavelet=Morlet()):
    cwta = lambda x: cwt_a(x, scales, dt, wavelet)
    return coherence_a(cwta(sig1), cwta(sig2), scales, dt, wavelet)


def wtc_f(sig1, sig2, freqs, Fs=1.0, wavelet=Morlet()):
    cwtf = lambda x: cwt_f(x, freqs, Fs, wavelet)
    return coherence_f(cwtf(sig1), cwtf(sig2), freqs, Fs, wavelet.f0)

def coherence_a(x,y,scales,dt,f0=1.5):
    # almost not useful 
    #scor =numpy.ones(x.shape[1])[:,numpy.newaxis] * scales
    #scor = numpy.transpose(scor)
    sx = wsmooth_a(abs(x)**2,scales,dt)
    sy= wsmooth_a(abs(y)**2,scales,dt)
    sxy = wsmooth_a((x*y.conjugate()), scales, dt)
    return abs(sxy)**2/(sx.real*sy.real)

def coherence_f(x,y,freqs, Fs=1.0, f0=1.5):
    return coherence_a(x,y,f0/freqs, 1.0/Fs, f0)

def cphase(x,y):
    #hard to interprete :(
    d = xwt(x,y)
    return numpy.arctan2(d.imag, d.real)

def cwt_phase(x):
    return numpy.arctan2(x.imag, x.real)


def wsmooth_a(coefs, scales, dt = 1.0, wavelet=Morlet()):
    "Smoothing of wavelet coefs. Scales version"
    #G = lambda omega,s: numpy.exp(-0.5 * s**2.0 *omega**2.0) # Literature
    G = lambda omega,s: numpy.exp(-0.25 * s**2.0 *omega**2.0) 

    W = numpy.zeros(coefs.shape, 'complex')
    fftom = 2*pi*fftfreq(coefs.shape[1], dt) 

    for n,s in enumerate(scales):
        W[n,:] = ifft(fft(coefs[n,:]) * G(fftom, s/dt))

    # TODO: scale smoothing
    return W





#!/usr/bin/env python

from distutils.core import setup

classifiers=[
      'Development Status :: 4 - Beta',
      'Environment :: X11 Applications :: GTK',
      'Intended Audience :: End Users/Desktop',
      'License :: OSI Approved :: GNU General Public License (GPL)',
      'Operating System :: POSIX :: Linux',
      'Programming Language :: Python',
      'Topic :: Scientific/Engineering',
        ]

setup(name='swan',
      version = "0.6.2",
      description = 'Easy continuous wavelet analysis',
      author = 'Alexey Brazhe',
      author_email = "brazhe@gmail.com",
      license = 'GNU GPL',
      url = "http://cell.biophys.msu.ru/static/swan",
      packages = ['swan', 'swan.gui'],
      package_dir = {'swan':'swan_modules'},
      package_data = {'swan.gui': ["glade/swan.glade"]},
      scripts = ['swan'],
      long_description = \
      """swan is a tool for wavelet data analysis. 
      It's meant to be simple in use and easy to extend.""",
      classifiers=classifiers,
      )




#!/usr/bin/env python
"""
Swan --- a tool for wavelet analysis of data, 
inspired by WAND, a Wavelet Analyser of Data by D. Postnov

I wanted to write my own version in python/GTK instead of TCL/Tk
and try to advance the ideas behind.
"""
#Copyright 2005-2008 Alexey Brazhe
#brazhe@biophys.msu.ru
#
#-----------------------------------------------------
# This file is part of Swan package.
#
# Swan is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# Swan is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# the software; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
# Suite 330, Boston, MA 02111-1307 USA 
#-----------------------------------------------------

# TODO: ...

try:
    import psyco
    psyco.full()
except:
    pass


import os, sys
import gtk
import gtk.glade
#from gzip import GzipFile as gzfile

import numpy

from matplotlib.figure import Figure

try:
    from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg as FigCanv
except ImportError:
    print "No gtkagg module, loading gtk"
    from matplotlib.backends.backend_gtk import FigureCanvasGTK as FigCanv

from matplotlib.backends.backend_gtk import NavigationToolbar2GTK as NToolbar

#from pylab import cm, rc, setp, load
from pylab import *


from swan import pycwt
from swan.gui import Rhythm_path

from swan.utils import *

def defined(obj):
    """ Unifies some tests """
    return (obj is not None)



def with_data(method):
    """Only do something when data have been loaded (decorator)"""
    def _(self,*args,**kwargs):
        if defined(self.data):
            method(self,*args,**kwargs)
        else:
            print "Have no data loaded, nothing to do..."
    return _

def with_cwt(method):
    """Only do something when cwt have been computed (decorator)"""
    def _(self, *args, **kwargs):
        if defined(self.cwt):
            method(self,*args,**kwargs)
        else:
            print "No cwt computed, nothing to do..."
    return _



from swancmap import set_rgbswan_cm, rgbswan

def determine_path ():
    """Borrowed from wxglade.py"""
    try:
        root = __file__
        if os.path.islink (root):
            root = os.path.realpath (root)
        return os.path.dirname (os.path.abspath (root))
    except:
        print "I'm sorry, but something is wrong."
        print "There is no __file__ variable. Please contact the author."
        sys.exit ()


class Swan:
    def __init__(self,fname=None,
                 Fs = 2.0,
                 f0 = 1.5,
                 labelsize = 10.0,
                 verbose = 1):
        
        self.verbose = verbose
        self.deoutbf = True # wether to remove outbursts

        self.data = None
        self.Fs = Fs # sampling frequency

        self.rc = {'glade-file':'/glade/swan.glade',
                   'host-path': determine_path(),
                   'label-size': labelsize,
                   'wavelet' : 'Morlet',
                   }

        self.wTree = gtk.glade.XML(self.rc['host-path'] + self.rc['glade-file'])
        self.window = self["swan_main"]
 
        self.fig = Figure(figsize=(8,6), dpi=100)

        self['morlet_f0'].set_value(f0)
        self['Fs_spin'].set_value(Fs)

        self.f_pars = {'start': 0.1, 'step': 0.025, 'stop': 10., 'log?': False} #frequency parameters
        #self.set_freqs = self.set_freqs_linear

        self.skeletons = []
        self.skel_plots = []
        
        self.rind = -1
        self.rps = []   # Rhythm paths

        self.cwt_freshly_runf = False
        self.ehndl = None # mouse click event handler
        self.cwt = None
        self.show_how = self.show_surface
        self.wavelet = pycwt.Morlet(f0)
        self.cwt_image = None

        # -- Prepare different parts --
        set_rgbswan_cm()
        self.setup_callbacks()
        self.setup_show_funcs()
        self.setup_colormaps()
        self.setup_axes()
        self.setup_canvas()
        self.setup_combos()


        # -- Try to load data if the file name is provided --
        if defined(fname):
            self.load_data(fname)
            self.plot_data()

        self.window.show_all()

    def setup_callbacks(self):
        wdict = {"on_load_data_activate": self.on_load_data_activate,
                 "on_Fs_spin_value_changed": self.on_Fs_changed,
                 "on_apply_cwt_clicked": self.on_apply_cwt_clicked,
                 "on_morlet_f0_value_changed": self.on_f0_changed,
                 "on_add_rhythm_clicked": self.new_rhythm,
                 "on_rem_rhythm_clicked": self.del_rhythm,
                 "on_apply_rhythm_clicked": self.apply_path,
                 "on_save_rhythm_clicked": self.save_rhythm,
                 "on_colormaps_combo_changed" : self.change_cmap,
                 "on_display_combo_changed" : self.change_display,
                 "on_show_how_combo_changed": self.change_show_how,
                 "on_freq_log_scale_toggled": self.on_freq_log_scale_toggled,
                 "on_local_maxima_toggled" : self.on_local_maxima_toggled,
                 "on_logscale_image_toggled" : self.image_logscale_toggle,
                 "on_f_start_value_changed": lambda x : self.on_f_params_changed(x,'start'),
                 "on_f_step_value_changed": lambda x : self.on_f_params_changed(x,'step'),
                 "on_f_stop_value_changed": lambda x : self.on_f_params_changed(x,'stop'),
                 "on_swan_main_destroy":gtk.main_quit,
                 "on_quit1_activate":gtk.main_quit
                }
        self.wdict=wdict
        self.wTree.signal_autoconnect(wdict)

    def setup_show_funcs(self):
        """ Prepare functions for different show options """
        self.show_funcs = {
            'Modulus': lambda x: abs(x),
            'Phase': lambda x: arctan2(x.imag, x.real),
            'Real' : lambda x: x.real,
            'Imaginary': lambda x: x.imag,
            'EDS': self.eds,
            #'EDS': lambda x: (x.real**2. + x.imag**2.)/self.wavelet.fc,
            'EDS-A-1': lambda x: self.eds_a_helper(x,1),
            'EDS-A-2': lambda x: self.eds_a_helper(x,2)
                }
        self.image_logscale = False
        self.show_func = self.show_funcs['EDS'] # default view

    def eds(self, x):
        return pycwt.eds(x)/std(self.data)**2

    def eds_a_helper(self, x, alpha):
        """ Show EDS with ridge amplitude correction """
        #alpha = 1.9798196483 # magic alpha
        #coefs = self.freqs**alpha * 2.0/ pi
        coefs = (self.freqs**alpha) * 2.0/ pi
        eds = self.eds(x)
        if len(eds.shape) >1: 
            for i in xrange(eds.shape[1]):
                eds[:,i] *= coefs
        else:
            eds *= coefs
        #return eds/self.wavelet.fc
        return eds


    def setup_colormaps(self):
        """ Sets the avalable colormpas up """
        self.curr_cmap = cm.jet
        self.known_cmaps = ('rgbswan', 'autumn', 'bone', 'binary', 'cool', 'copper', 'flag',\
                'gray', 'hot', 'hsv', 'jet', 'pink', \
                'prism', 'spectral', 'spring', 'summer', 'winter')
        for m in self.known_cmaps:
            self['colormaps_combo'].append_text(m)
        #self['colormaps_combo'].set_active(8)

    def setup_combos(self):
        """ Sets default values in combo boxes"""
        self['display_combo'].set_active(1)
        self['show_how_combo'].set_active(0)
        self['colormaps_combo'].set_active(0)
    
    def load_data(self,fname):
        """ Loads the data """
        try:
            data = numpy.loadtxt(fname)
            s = data.shape
            self.fname = fname
            
            if len(s) < 2:
                if len(data)%2 :
                    data = data[:-1]
                    #self.data = data - mean(data)
                self.data = detrend1(data)
                if self.deoutbf:
                    data = deoutburst(data)
                    self.data = detrend1(data)

                self.time_from_Fs(self.Fs)
            else:
                if s[0] < s[1]:
                    data = transpose(data)
                    s = data.shape

                time = data[:,0] / 1000.0 # Here I assume that time is in ms
                data = data[:,1]
                
                if len(data)%2 :
                    data = data[:-1]
                    time = time[:-1]
                
                self.data = data
                self.time = time
                self.Fs = 1./mean(time[1:]- time[:-1])
                self['Fs_spin'].set_value(self.Fs)

        except:
            print "Error in data loading:"
            print sys.exc_type, sys.exc_value
            self.fname = None
            self.data = None

    def setup_canvas(self):
        """Prepares canvas to draw in"""
        self.canvas = FigCanv(self.fig)
        self.toolbar = NToolbar(self.canvas, None)
        self['pylab_box'].pack_start(self.canvas,True,True)
        self['pylab_box'].pack_start(self.toolbar, False, False)

    def setup_axes(self):
        """Prepares axes for further work"""
        rc('xtick', labelsize = self.rc['label-size'])
        rc('ytick', labelsize = self.rc['label-size'])
        self.fig.clear()
        self.ax = [self.fig.add_axes((0.04, 0.2, 0.8, 0.78))]
        self.ax.append(self.fig.add_axes((0.04, 0.04, 0.8, 0.12), sharex=self.ax[0]))
        self.ax.append(self.fig.add_axes((0.85, 0.2, 0.04, 0.78), 
                                         xticklabels=[], 
                                         yticklabels=[],
                                         visible = False))
        self.ax[0].hold(True)
        self.ax[1].hold(False)
        #self.ax[2].hold(False)

    def on_load_data_activate(self, widget):
        """ Callback for the Open menu item """
        fname = self.select_data_file()
        if fname is None:
            if self.verbose:
                print "We didn't choose any file "
            return
        if self.verbose:
            print "We chose the file %s" % fname
        self.load_data(fname)
        self.plot_data()


    def on_Fs_changed(self,widget):
        """ Callback to change the sampling frequency """
        self.Fs = widget.get_value()
        self.time_from_Fs(self.Fs)
        self.plot_data()
        return 0

    def on_f_params_changed(self, spinw, key):
        "Update  frequency parameters for cwt"
        if self.verbose >1:
            print 'Changing key:', key
        self.f_pars[key] = spinw.get_value()


    def on_freq_log_scale_toggled(self,w):
        self.f_pars['log?'] = w.get_active()


    def image_logscale_toggle(self,w):
        self.image_logscale = w.get_active();
        self.change_display(self['display_combo'])

    @with_data
    def on_apply_cwt_clicked(self, w):
        self.freqs = self.set_freqs()
        self.cwt_calc()
        self.cwt_display()


    def on_f0_changed(self,w):
        f0 = w.get_value()
        self.wavelet.set_f0(f0)
        self.show_funcs['EDS'] = lambda x: (x.real**2. + x.imag**2.)/self.wavelet.fc

    def set_freqs(self):
        if self.f_pars['log?']:
            a = log2(self.f_pars['start'])
            b = log2(self.f_pars['stop'])
            if self.verbose:
                print a,b, 1/self.f_pars['step']
            return 2.**arange(a, b, self.f_pars['step'])
        else:
            return arange(self.f_pars['start'],
                          self.f_pars['stop'],
                          self.f_pars['step'])

        return 2.**arange(a, b, self.f_pars['step'])

    @with_data
    def cwt_calc(self):
        self.cwt = pycwt.cwt_f(self.data, self.freqs, Fs=self.Fs, \
                wavelet = self.wavelet)
        self.cwt_freshly_runf = True
        pass

    @with_cwt
    def cwt_display(self):
        """Show the results of the wavelet transform"""
        
        self.clear_all_rhythms()
        self.ax[0].clear()
        self.rm_skeletons()
        self['local_maxima_chk'].set_active(False)

        self.show_how()
        self.show_cone()
        setp(self.ax[0],'yscale',('linear','log')[self.f_pars['log?']])
        if self.ehndl is None:
            self.ehndl = self.canvas.mpl_connect('button_press_event',self.cont_path)
        self.show_colorbar()
        self.canvas.draw()

    @with_cwt
    def show_cone(self):
        w = self.wavelet
        f = self.freqs
        ax = self.ax[0]
        t = self.time[-1]
        try:
            ax.fill_betweenx(f, 0.0,
                             w.cone(f),
                             alpha=0.5, color='black')
            ax.fill_betweenx(f,
                             t+w.cone(-f),
                             t,
                             alpha=0.5, color='black')
        except:
            print("Can't use fill_betweenx function: update\
            maptlotlib?")


    
    
    @with_cwt
    def show_surface(self):
        self.cwt_image = self.ax[0].imshow(self.show_func(self.cwt),\
                                               origin='lower',\
                                               extent=(self.time[0], self.time[-1], \
                                                           self.freqs[0], self.freqs[-1]),\
                                               interpolation='bilinear',\
                                               cmap=self.curr_cmap)



    @with_cwt
    def show_contour(self):
        self.cwt_image = self.ax[0].contour(self.show_func(self.cwt), \
                                                origin='lower',\
                                                extent=(self.time[0], self.time[-1], \
                                                            self.freqs[0], self.freqs[-1]),\
                                                cmap=self.curr_cmap)
    @with_cwt
    def show_colorbar(self):
        self.fig.colorbar(self.cwt_image, cax=self.ax[2])
        setp(self.ax[2], visible=True)

    def change_show_how(self, w):
        """ Callback for the 'show how' combo"""
        val = w.get_active_text()
        if val == 'Surface':
            self.show_how = self.show_surface
        else:
            self.show_how = self.show_contour
        self.cwt_display()
        pass


    def select_data_file(self, data=None):
        dlg = gtk.FileChooserDialog('Select data file',
                                    None,
                                    action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                    buttons=(gtk.STOCK_CANCEL,
                                             gtk.RESPONSE_CANCEL,
                                             gtk.STOCK_OPEN,
                                             gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        flt = gtk.FileFilter()
        flt.set_name('Data files')
        flt.add_pattern('*.*')
        dlg.add_filter(flt)
        resp = dlg.run()
        if resp ==gtk.RESPONSE_OK:
            fname = dlg.get_filename()
        else:
            fname = None
        dlg.destroy()
        return fname

    @with_data
    def time_from_Fs(self,Fs=2.0):
        tfinish = len(self.data)/Fs
        self.time = arange(0,1.2*tfinish,1./Fs)[:len(self.data)]



    @with_data
    def plot_data(self):
        """Plot current data"""
        if self.time is None:
            self.ax[1].plot(self.data,'k-')
        else:
            self.ax[1].plot(self.time,self.data,'k-')
            setp(self.ax[1],'xlim',(self.time[0], self.time[-1]))
        self.canvas.draw()
        return 0

    def on_local_maxima_toggled(self,w,data=None):
        """Show on/off the skeletons line"""
        #TODO: re-structure this!
        if w.get_active():
            if self.renew_skelsp():
                self.calc_skeletons()
                self.skel_plots = []
                for skel in self.skeletons:
                    self.skel_plots.append(self.plot_a_skeleton(skel))
                self.set_ax0_tight()
                self.canvas.draw()
            else:
                self.skeletons_shade(False)
        else:
            self.skeletons_shade(True)

    def renew_skelsp(self):
        return defined(self.data) and \
                (len(self.skeletons) is 0) and\
                self.cwt_freshly_runf

    
    def plot_a_skeleton(self,skel):
        return self.ax[0].plot(skel[:,0], skel[:,1], 'y.', markersize = 0.1)

    def set_ax0_tight(self):
        setp(self.ax[0], 'xlim', (0, self.time[-1]),\
                'ylim', (self.freqs.min(), self.freqs.max()))
    
    def skeletons_shade(self, flag=True):
        for plh in self.skel_plots:
            setp(plh,'visible', not flag)
        self.canvas.draw()

    def rm_skeletons(self):
        self.skeletons_shade()
        del(self.skeletons[:])
        del(self.skel_plots[:])
        pass

    def calc_skeletons(self):
        """calculate local modulus maxima (skeleton lines)"""
        L =  len(self.data)
        del(self.skeletons[:])
        for i in xrange(L):
            xe,ye = get_extr_deriv(self.freqs, self.show_func(self.cwt[:,i]))
            ye[:] = self.time[i]
            self.skeletons.append(transpose(array([ye,xe])))
            if not i%10:
                sys.stderr.write (make_progress_str(i,L))

    def change_cmap(self,w,udata=None):
        """Change a color map"""
        cmkey = w.get_active_text()
        self.curr_cmap = getattr(cm, cmkey)
        if defined(self.cwt_image):
            setp(self.cwt_image,cmap=self.curr_cmap)
        self.canvas.draw()

    
    def change_display(self,w,udata = None):
        """Change display mode"""
        display_key = w.get_active_text()
        if display_key is None:
            display_key = 'Modulus'
        show_func = self.show_funcs[display_key]
        if self.image_logscale and ((display_key == 'Modulus') or 
                                    (display_key == 'EDS') or
                                    (display_key == 'EDS-A-1') or 
                                    (display_key == 'EDS-A-2')):
            self.show_func = lambda x: log(show_func(x))
        else:
            self.show_func = show_func
        self.cwt_display()


    def toggle_path(self,w,id=0,but=1):
        """Show on/off for the selected path or rhythm"""
        if self.verbose > 1:
            print 'Caller: %d, %s' % (id, ('P', 'R')[but-1])
        if but == 1:
            self.rps[id].preplot(self.canvas,w.get_active())
        elif but ==2:
            self.rps[id].rreplot(self.canvas,w.get_active())
        
    def  toggle_rhythm(self,w,data=0):
        """Select on rhythm-path and deselect others"""
        if self.verbose > 1:
            print 'Caller: %d' % data
        for rp in self.rps:
            if rp.id == data:
                self.rind = (-1,data)[w.get_active()]
                if self.verbose >2 :
                    print 'self.rind ', self.rind
            else:
                rp.buts[0].handler_block(rp.butcb[0])      # fantastiske :)
                rp.buts[0].set_active(False)
                rp.buts[0].handler_unblock(rp.butcb[0])    # fantastiske :)


    def clear_all_rhythms(self):
        while len(self.rps) > 0:
            self.del_rhythm()

    def del_rhythm(self,w=None,data=None):
        x = self.rps.pop()
        x.__del__()
        if self.rind > -1:
            self.rind -= 1
        self.window.show_all()

    def new_rhythm(self,w, data=None):
        """Starts new rhythm path"""
        i = len(self.rps)
        nrp = Rhythm_path.Rhythm_path(self.canvas, i, 0.04)
        nrp.butcb[0] = nrp.buts[0].connect('toggled',self.toggle_rhythm,i)
        nrp.buts[0].set_active(True)
        nrp.buts[1].set_active(True)
        nrp.butcb[1] = nrp.buts[1].connect('toggled',self.toggle_path,i,1)
        nrp.butcb[2] = nrp.buts[2].connect('toggled',self.toggle_path,i,2)
        #self.boxes['ctrls/rh'].pack_start(nrp.box,False,False,0)
        self['rhythms_store'].pack_start(nrp.box,False,False,0)
        nrp.plot_setup(self.ax[0],[(0,self.time[-1]), (self.freqs[0], self.freqs[-1])])
        self.rps.append(nrp)
        self.rind = i
        self.window.show_all()

    def cont_path(self,event):
        """Continues the rhythm path by adding points or removing them"""
        i = self.rind
        if i > -1:
            if event.inaxes == self.ax[0]:
                crps = self.rps[i]
                if event.button == 1:                      # left click
                    crps.push_point(event.xdata,event.ydata)
                elif event.button == 3:                    # right click
                    crps.pop_point(event.xdata,event.ydata)
                elif event.button == 2:                    # middle click
                    crps.clear()
                crps.preplot(self.canvas)

    def apply_path(self,w,udata=None):
        i = self.rind
        if i > -1:
            self.path_2_rhythm0()
            crps = self.rps[i]
            if crps.buts[2].get_active():
                crps.rreplot(self.canvas)
            else:
                crps.buts[2].set_active(True)

    def path_2_rhythm0(self):

        i = self.rind                   # index of the current rhythm path
        crps = self.rps[i]              # current rhythm path
        
        x,y,v = [],[],[]
        px,py,pu,pl = crps.vectors_from_bpoints()
        
        L = self.get_xind(px[-1]) - self.get_xind(px[0]) + 1
        
        for k in range(len(px) - 1):
            p = polyfit(px[k:k+2], py[k:k+2], 1)  # Linear approximation between two path points
            sy = arange(self.cwt.shape[0])            # eds surface y-indices
            jstart = self.get_xind(px[k])
            jstop = self.get_xind(px[k+1])
            self.ax[0].hold(True)
            for j in range(jstart,jstop):
                yslice = self.show_func(self.cwt[:,j])
                ym = p[0]*self.time[j] + p[1]
                yl = min(sy[self.freqs > ym-crps.band]) 
                yu = max(sy[self.freqs <= (ym + crps.band)])+1
                tmp = argsort(yslice[yl:yu])[-1]
                x.append(self.time[j])
                v.append(yslice[yl+tmp])
                y.append(self.freqs[yl+tmp])
                if self.verbose:
                    if not j % 10:
                        sys.stderr.write (make_progress_str(j,L))
        crps.x = x
        crps.y = y
        crps.v = v
        if self.verbose:
            print '\nDone with path to rhythm conversion'
    


    def save_rhythm(self,w,udata=None):
        i = self.rind
        if len(self.rps[i].x) > 0:
            ya = zeros((len(self.rps[i].x)),dtype='d')
            fname = self.select_rhythm()
            if defined(fname):
                out_f = fname + '_f.dat'  
                out_a = fname + '_a.dat'  
                save(out_f,self.rps[i].y)
                save(out_a,self.rps[i].v)
                print fname
            

    def select_rhythm(self, data=None):
        dlg = gtk.FileChooserDialog('Where should we save our rhythm?',
                None,
                action=gtk.FILE_CHOOSER_ACTION_SAVE,
                buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_SAVE,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        flt = gtk.FileFilter()
        flt.set_name('All files')
        flt.add_pattern('*')
        dlg.add_filter(flt)
        resp = dlg.run()
        if resp ==gtk.RESPONSE_OK:
            fname = dlg.get_filename()
        else:
            fname = None
        dlg.destroy()
        return fname


    def get_xind(self,xp):
        return int(around(xp*self.Fs))

                
    def main(self):
        gtk.main()
    
    def __getitem__(self,key):
        return self.wTree.get_widget(key)
#------------------- End of the Swan class ------------

